@@ -48,7 +48,7 @@ extern "C" {
 /**************************************************/
 
 
-/* Format and for printf-compatibel format-strings
+/* Format and for printf-compatible format-strings
  * example use: printf("physical address: %" IMX_VPU_PHYS_ADDR_FORMAT, phys_addr */
 #define IMX_VPU_PHYS_ADDR_FORMAT "#lx"
 /* Typedef for physical addresses */
@@ -72,9 +72,9 @@ ImxVpuAllocationFlags;
  * both be set */
 typedef enum
 {
-	/* Map memory for CPU read access */
-	IMX_VPU_MAPPING_FLAG_WRITE   = (1UL << 0),
 	/* Map memory for CPU write access */
+	IMX_VPU_MAPPING_FLAG_WRITE   = (1UL << 0),
+	/* Map memory for CPU read access */
 	IMX_VPU_MAPPING_FLAG_READ    = (1UL << 1)
 	/* XXX: When adding extra flags here, follow the pattern: IMX_VPU_MAPPING_FLAG_<NAME> = (1UL << <INDEX>) */
 }
@@ -311,17 +311,15 @@ typedef enum
 {
 	/* MPEG-1 part 2 and MPEG-2 part 2.
 	 *
-	 * Decoding: Fully compatible with the ISO/IEC 13182-2 specification and
-	 * the main and high profiles. Both progressive and interlaced content is
-	 * supported. */
+	 * Decoding: Fully compatible with the ISO/IEC 13182-2 specification and the main and high
+	 *           profiles. Both progressive and interlaced content is supported.
+	 */
 	IMX_VPU_CODEC_FORMAT_MPEG2 = 0,
 
 	/* MPEG-4 part 2.
 	 *
 	 * Decoding: Supports simple and advanced simple profile (except for GMC).
-	 * NOTE: DivX 3/5/6 are not supported and require special licensing by
-	 * Freescale.
-	 *
+	 *           NOTE: DivX 3/5/6 are not supported and require special licensing by Freescale.
 	 * Encoding: Supports the simple profile and max. level 5/6.
 	 */
 	IMX_VPU_CODEC_FORMAT_MPEG4,
@@ -336,14 +334,17 @@ typedef enum
 	/* h.264.
 	 *
 	 * Decoding: Supports baseline, main, high profiles, max. level 4.1.
+	 *           (10-bit decoding is not supported.)
+	 *           Only Annex.B byte-stream formatted input is supported.
 	 * Encoding: Supports baseline and constrained baseline profile, max. level 4.0.
+	 *           Only Annex.B byte-stream formatted output is supported.
 	 */
 	IMX_VPU_CODEC_FORMAT_H264,
 
 	/* WMV3, also known as Windows Media Video 9. Compatible to VC-1 simple and main profiles.
 	 *
-	 * Decoding: Fully supported WMV3 decoding, excluding the deprecated WMV3
-	 * interlace support (which has been obsoleted by the interlacing in the
+	 * Decoding: Fully supported WMV3 decoding, excluding the deprecated WMV3 interlace support
+	 *           (which has been obsoleted by the interlacing in the VC-1 advanced profile).
 	 * VC-1 advanced profile). */
 	IMX_VPU_CODEC_FORMAT_WMV3,
 
@@ -356,19 +357,18 @@ typedef enum
 	/* Motion JPEG.
 	 *
 	 * Decoding: Only baseline JPEG frames are supported. Maximum resolution is 8192x8192.
-	 *
 	 * Encoding: Only baseline JPEG frames are supported. Maximum resolution is 8192x8192.
-	 * MJPEG always operates in constant quality mode, even if the encoder open params
-	 * have a nonzero bitrate set.
+	 *           NOTE: Encoder always operates in constant quality mode, even if the open
+	 *           params have a nonzero bitrate set.
 	 */
 	IMX_VPU_CODEC_FORMAT_MJPEG,
 
 	/* VP8.
 	 *
-	 * Decoder: fully compatible with the VP8 decoding specification.
-	 * Both simple and normal in-loop deblocking are supported.
-	 * NOTE: VPU specs state that the maximum supported resolution is 1280x720, but tests
-	 * show that up to 1920x1088 pixels do work.
+	 * Decoding: fully compatible with the VP8 decoding specification.
+	 *           Both simple and normal in-loop deblocking are supported.
+	 *           NOTE: VPU specs state that the maximum supported resolution is 1280x720, but
+	 *           tests show that up to 1920x1088 pixels do work.
 	 */
 	IMX_VPU_CODEC_FORMAT_VP8
 
@@ -618,6 +618,8 @@ char const *imx_vpu_frame_type_string(ImxVpuFrameType frame_type);
  *    convenience function for this is strongly recommended.
  * 9. Call imx_vpu_dec_register_framebuffers() and pass in the ImxVpuFramebuffer array
  *    and the number of ImxVpuFramebuffer instances.
+ *    Note that this call does _not_ copy the framebuffer array, it just stores the pointer
+ *    to it internally, so make sure the array is valid until the decoder is closed!
  *    This should be the last action in the callback.
  * 10. Continue calling imx_vpu_dec_decode(). Make sure the input data is not NULL.
  *     If the IMX_VPU_DEC_OUTPUT_CODE_DECODED_FRAME_AVAILABLE flag is set in the output code,
@@ -652,18 +654,20 @@ char const *imx_vpu_frame_type_string(ImxVpuFrameType frame_type);
  *
  *   mutex_lock(&mutex);
  *
- *   while (!imx_vpu_dec_check_if_can_decode(decode) && !abort_waiting)
+ *   while (dec_initialized && !imx_vpu_dec_check_if_can_decode(decode) && !abort_waiting)
  *     condition_wait(&condition_variable, &mutex);
  *
  *   if (!abort_waiting)
- *     imx_vpu_dec_decode_frame(decoder, encoded_frame, &output_code);
+ *     imx_vpu_dec_decode(decoder, encoded_frame, &output_code);
  *   ...
  *
  *   mutex_unlock(&mutex);
  *
  * (abort_waiting would be a flag that gets raised when something from the outside signals
  * that waiting and decoding needs to be shut down now, for example because the user wants
- * to close the player, or because the user pressed Ctrl+C.)
+ * to close the player, or because the user pressed Ctrl+C. dec_initialized would be a flag
+ * that is initially cleared, and raised in the initial info callback; it is pointless to
+ * call imx_vpu_dec_check_if_can_decode() before the callback was executed.)
  *
  * If any video sequence parameters (like frame width and height) in the input data change,
  * the output code from imx_vpu_dec_decode() calls in step 10 will contain the
@@ -740,7 +744,11 @@ typedef enum
 	 * to a following imx_vpu_dec_decode() call; instead, new data
 	 * should be loaded. If this code is not present, then the decoder
 	 * didn't use it yet, so give it to the decoder again until this
-	 * code is set or an error is returned. */
+	 * code is set or an error is returned.
+	 * NOTE: this flag is obsolete. It used to mean something with the
+	 * fslwrapper backend; however, with the vpulib backend, it will
+	 * always use the input unless an error occurs or EOS is signaled
+	 * in drain mode. */
 	IMX_VPU_DEC_OUTPUT_CODE_INPUT_USED                   = (1UL << 0),
 	/* EOS was reached; no more unfinished frames are queued internally.
 	 * This can be reached either by bitstreams with no frame delay,
@@ -767,8 +775,12 @@ typedef enum
 	IMX_VPU_DEC_OUTPUT_CODE_NOT_ENOUGH_INPUT_DATA        = (1UL << 5),
 	/* The VPU detected a change in the video sequence parameters
 	 * (like frame width and height). Decoding cannot continue. See the
-	 * explanation in th step-by-step guide above for what steps to take
-	 * if this output code is set. */
+	 * explanation in the step-by-step guide above for what steps to take
+	 * if this output code is set. Note that this refers to detected
+	 * changes in the *input data*, not to the decoded frames. This means
+	 * that this flag is set immediately when input data with param changes
+	 * is fed to the decoder, even if this is for example a h.264 high
+	 * profile stream with lots of frame reordering and frame delays. */
 	IMX_VPU_DEC_OUTPUT_CODE_VIDEO_PARAMS_CHANGED         = (1UL << 6)
 }
 ImxVpuDecOutputCodes;
@@ -892,10 +904,18 @@ ImxVpuDecReturnCodes imx_vpu_dec_flush(ImxVpuDecoder *decoder);
  * set in it. Registering can happen only once during the lifetime of a decoder instance. If for some reason
  * framebuffers need to be re-registered, the instance must be closed, and a new one opened.
  * The caller must ensure that the specified framebuffer array remains valid until the decoder instance
- * is closed. Note that internally, values might be written to the array (though it will never be reallocated
+ * is closed, since this function does not copy it; it just stores a pointer to the array internally. Also
+ * note that internally, values might be written to the array (though it will never be reallocated
  * and/or freed from the inside). Also, the framebuffers' DMA buffers will be memory-mapped until the decoder
  * is closed.
  *
+ * Since this function only stores a pointer to the framebuffer array internally, and does not actually copy
+ * the array, it is possible - and valid - to modify the "context" fields of the framebuffers even after
+ * this call was made. This is useful if for example system resources are associated later with the
+ * framebuffers. In this case, it is perfectly OK to set "context" to NULL initially, and later, when the
+ * resources are available, associated them to the framebuffers by setting the context fields, even if
+ * imx_vpu_dec_register_framebuffers() was already called earlier.
+ *
  * The framebuffers must contain valid values. The convenience functions imx_vpu_calc_framebuffer_sizes() and
  * imx_vpu_fill_framebuffer_params() can be used for this. Note that all framebuffers must have the same
  * stride values. */
@@ -937,7 +957,8 @@ void imx_vpu_dec_get_dropped_frame_info(ImxVpuDecoder *decoder, void **context,
  * to decode. This is directly related to the set of free framebuffers. If this function returns 0, decoding
  * should not be attempted until after imx_vpu_dec_mark_framebuffer_as_displayed() was called. If this
  * happens, imx_vpu_dec_check_if_can_decode() should be called again to check if the situation changed and
- * decoding can be done again. See the explanation above for details. */
+ * decoding can be done again. Also, calling this function before the initial info callback was executed is
+ * not recommended and causes undefined behavior. See the explanation above for details. */
 int imx_vpu_dec_check_if_can_decode(ImxVpuDecoder *decoder);
 
 /* Marks a framebuffer as displayed. This always needs to be called once the application is done with a decoded
@@ -1018,11 +1039,14 @@ ImxVpuDecReturnCodes imx_vpu_dec_mark_framebuffer_as_displayed(ImxVpuDecoder *de
  *     to set default values. It is essential to make sure the acquire_output_buffer() and
  *     finish_output_buffer() function pointers are set, as these are used for acquiring buffers
  *     to write encoded output data into.
+ *     Alternatively, set write_output_data() if write-callback style output is preferred. If this
+ *     function pointer is non-NULL, then acquire_output_buffer() and finish_output_buffer() are
+ *     ignored.
  * 13. If step 8 was performed, and therefore input data does *not* come in DMA memory, copy the
  *     pixels from the raw input frames into the DMA buffer allocated in step 8. Otherwise, if
  *     the raw input frames are already stored in DMA memory, this step can be omitted.
  * 14. Call imx_vpu_enc_encode(). Pass the raw frame, the encoded frame, and the encoding param
- *     strucutres from steps 9, 10, and 12 to it.
+ *     structures from steps 9, 10, and 12 to it.
  *     This function will encode data, and acquire an output buffer to write the encoded data into
  *     by using the acquire_output_buffer() function pointer set in step 12. Once it is done
  *     encoding, it will call the finish_output_buffer() function from step 12. Any handle created
@@ -1037,6 +1061,10 @@ ImxVpuDecReturnCodes imx_vpu_dec_mark_framebuffer_as_displayed(ImxVpuDecoder *de
  *     actual encoded frame data. imx_vpu_enc_encode() will pass over the combined size of the header
  *     and the encoded frame data to acquire_output_buffer() in this case, ensuring that the output
  *     buffers are big enough.
+ *     If write-callback style output is used instead (= if the write_output_data() function pointer
+ *     inside the encoding_params is set to a valid value), then this function haves as described
+ *     above, except that it does not call acquire_output_buffer() or finish_output_buffer(). It
+ *     still adds headers etc. but outputs these immediately by calling write_output_data().
  * 15. Repeat steps 11 to 14 until there are no more frames to encode or an error occurs.
  * 16. After encoding is finished, close the encoder with imx_vpu_enc_close().
  * 17. Deallocate framebuffer memory blocks, the input DMA buffer block, the output memory block,
@@ -1045,6 +1073,15 @@ ImxVpuDecReturnCodes imx_vpu_dec_mark_framebuffer_as_displayed(ImxVpuDecoder *de
  * Note that the encoder does not use any kind of frame reordering. h.264 data uses the
  * baseline profile. An input frame immediately results in an output frame (unless an error occured).
  * There is no delay.
+ *
+ * The VPU's encoders only support the IMX_VPU_COLOR_FORMAT_YUV420 format, with the exception of
+ * the MJPEG encoder, which supports all formats from ImxVpuColorFormat. However, it is possible
+ * to provide YUV400 grayscale encoding support to the other encoders by internally using YUV420
+ * and using two dummy U and V planes that are filled with 0x80 bytes. For this purpose, the
+ * additional_dmabuffers_allocator from ImxVpuEncOpenParams is used. Framebuffers however do not
+ * have to allocate any space for dummy UV planes - the Y plane suffices.
+ * Note though that this trick, this "fake grayscale mode", will technically produce YUV420 data,
+ * and will be decoded by players as such, and NOT as YUV400 video.
  */
 
 
@@ -1079,7 +1116,9 @@ typedef enum
 	/* A function was called at an inappropriate time. */
 	IMX_VPU_ENC_RETURN_CODE_WRONG_CALL_SEQUENCE,
 	/* The operation timed out. */
-	IMX_VPU_ENC_RETURN_CODE_TIMEOUT
+	IMX_VPU_ENC_RETURN_CODE_TIMEOUT,
+	/* write_output_data() in ImxVpuEncParams returned 0. */
+	IMX_VPU_ENC_RETURN_CODE_WRITE_CALLBACK_FAILED
 }
 ImxVpuEncReturnCodes;
 
@@ -1148,6 +1187,17 @@ typedef enum
 ImxVpuEncMESearchRanges;
 
 
+typedef enum
+{
+	IMX_VPU_ENC_HEADER_DATA_TYPE_H264_SPS_RBSP = 0,
+	IMX_VPU_ENC_HEADER_DATA_TYPE_H264_PPS_RBSP,
+	IMX_VPU_ENC_HEADER_DATA_TYPE_MPEG4_VOS,
+	IMX_VPU_ENC_HEADER_DATA_TYPE_MPEG4_VIS,
+	IMX_VPU_ENC_HEADER_DATA_TYPE_MPEG4_VOL
+}
+ImxVpuEncHeaderDataTypes;
+
+
 /* Slice mode information to be used when opening an encoder instance. */
 typedef struct
 {
@@ -1277,7 +1327,9 @@ typedef struct
 	 * Default value is 16. */
 	unsigned int gop_size;
 	/* Color format to use for incoming frames. Only MJPEG actually uses
-	 * this value; other codec formats always use IMX_VPU_COLOR_FORMAT_YUV420.
+	 * all possible values; other codec formats only allow for the two formats
+	 * IMX_VPU_COLOR_FORMAT_YUV420 and IMX_VPU_COLOR_FORMAT_YUV400 (the second
+	 * one is supported by using YUV420 and dummy U and V planes internally).
 	 * See the ImxVpuColorFormat documentation for an explanation how
 	 * the chroma_interleave value can affec the pixel format that is used. */
 	ImxVpuColorFormat color_format;
@@ -1359,6 +1411,11 @@ typedef struct
 	 * plane, otherwise they are separated in their own planes.
 	 * See the ImxVpuColorFormat documentation for the consequences of this. */
 	int chroma_interleave;
+
+	/* Allocator for additional internal DMA buffers, such as the dummy U and V
+	 * planes for the fake YUV400 encoding (when MJPEG is not the codec format).
+	 * If this is NULL, imx_vpu_enc_get_default_allocator() is used. */
+	ImxVpuDMABufferAllocator *additional_dmabuffers_allocator;
 }
 ImxVpuEncOpenParams;
 
@@ -1390,6 +1447,8 @@ ImxVpuEncInitialInfo;
  * of ImxVpuEncodedFrame.
  * The return value is a pointer to a memory-mapped region of the
  * output buffer, or NULL if acquiring failed.
+ * If the write_output_data function pointer in the encoder params
+ * is non-NULL, this function is not called.
  * This function is only used by imx_vpu_enc_encode(). */
 typedef void* (*ImxVpuEncAcquireOutputBuffer)(void *context, size_t size, void **acquired_handle);
 /* Function pointer used during encoding for notifying that the encoder
@@ -1397,8 +1456,19 @@ typedef void* (*ImxVpuEncAcquireOutputBuffer)(void *context, size_t size, void *
  * allocated buffers; instead, it makes it possible to release, unmap etc.
  * context is the value of output_buffer_context specified in
  * ImxVpuEncParams. acquired_handle equals the value of *acquired_handle in
- * ImxVpuEncAcquireOutputBuffer. */
+ * ImxVpuEncAcquireOutputBuffer.
+ * If the write_output_data function pointer in the encoder params
+ * is non-NULL, this function is not called. */
 typedef void (*ImxVpuEncFinishOutputBuffer)(void *context, void *acquired_handle);
+/* Function pointer used during encoding for passing the output encoded data
+ * to the user. If this function is not NULL, then ImxVpuEncFinishOutputBuffer
+ * and ImxVpuEncAcquireOutputBuffer function are not called. Instead, this
+ * data write function is called whenever the library wants to write output.
+ * encoded_frame contains valid pts, dts, and context data which was copied
+ * over from the corresponding raw frame.
+ * Returns 1 if writing succeeded, 0 otherwise.
+ * */
+typedef int (*ImxVpuWriteOutputData)(void *context, uint8_t const *data, uint32_t size, ImxVpuEncodedFrame *encoded_frame);
 
 
 typedef struct
@@ -1421,9 +1491,25 @@ typedef struct
 	/* Functions for acquiring and finishing output buffers. See the
 	 * typedef documentations above for details about how these
 	 * functions should behave, and the imx_vpu_enc_encode()
-	 * documentation for how they are used. */
+	 * documentation for how they are used.
+	 * Note that these functions are only used if write_output_data
+	 * is set to NULL.
+	 */
 	ImxVpuEncAcquireOutputBuffer acquire_output_buffer;
 	ImxVpuEncFinishOutputBuffer finish_output_buffer;
+
+	/* Function for directly passing the output data to the user
+	 * without copying it first.
+	 * Using this function will inhibit calls to acquire_output_buffer
+	 * and finish_output_buffer. See the typedef documentations
+	 * above for details about how this function should behave, and
+	 * the imx_vpu_enc_encode() documentation for how they are used.
+	 * Note that if this function is NULL then acquire_output_buffer
+	 * and finish_output_buffer must be set.
+	 */
+	ImxVpuWriteOutputData write_output_data;
+
+	/* User supplied value that will be passed to the functions */
 	void *output_buffer_context;
 
 	/* Quantization parameter. Its value and valid range depends on
@@ -1487,9 +1573,30 @@ ImxVpuEncReturnCodes imx_vpu_enc_flush(ImxVpuEncoder *encoder);
  * the "min_num_required_framebuffers" field of ImxVpuEncInitialInfo. */
 ImxVpuEncReturnCodes imx_vpu_enc_register_framebuffers(ImxVpuEncoder *encoder, ImxVpuFramebuffer *framebuffers, unsigned int num_framebuffers);
 
-/* Retrieves initial information available after calling imx_vpu_enc_open(). */
+/* Retrieves initial information available after calling imx_vpu_enc_open().
+ * Internally this also generates stream headers (SPS/PPS RBSP for h.264,
+ * VOS/VIS/VOL headers for MPEG4). */
 ImxVpuEncReturnCodes imx_vpu_enc_get_initial_info(ImxVpuEncoder *encoder, ImxVpuEncInitialInfo *info);
 
+/* Retrieves a const pointer to a header, as well as that header's size in bytes.
+ * Do not try to modify the data the pointer refers to. To modify heade data, use
+ * imx_vpu_enc_set_header_data() instead.
+ * Using headers that belong to one format even though the encoder is configured
+ * for another format (for example, querying h.264 SPS RBSP even though MPEG-4 is
+ * configured) leads to undefined behavior.
+ * Header data will not be available before the imx_vpu_enc_get_initial_info() call.*/
+void imx_vpu_enc_query_header_data(ImxVpuEncoder *encoder, ImxVpuEncHeaderDataTypes header_data_type, uint8_t const **header_data, size_t *header_data_size);
+/* Sets new header data. Useful for inserting additional information into the
+ * headers. The data pointer to by the header_data pointer is copied to an
+ * internally allocated buffer. Any previously existing header data buffer will
+ * first be deallocated.
+ * If allocation fails, this returns IMX_VPU_ENC_RETURN_CODE_ERROR. Do not try to
+ * further use the encoder, since header data is missing. Otherwise, this returns
+ * IMX_VPU_ENC_RETURN_CODE_OK.
+ * Set this only after the imx_vpu_enc_get_initial_info() call, otherwise the header
+ * data will be overwritten. */
+ImxVpuEncReturnCodes imx_vpu_enc_set_header_data(ImxVpuEncoder *encoder, ImxVpuEncHeaderDataTypes header_data_type, uint8_t const *header_data, size_t header_data_size);
+
 /* Set the fields in "encoding_params" to valid defaults
  * Useful if the caller wants to modify only a few fields (or none at all) */
 void imx_vpu_enc_set_default_encoding_params(ImxVpuEncoder *encoder, ImxVpuEncParams *encoding_params);
@@ -1516,7 +1623,7 @@ void imx_vpu_enc_configure_intra_qp(ImxVpuEncoder *encoder, int intra_qp);
  * (which is given to acquire_output_buffer() as an argument). The return value of acquire_output_buffer()
  * is a pointer to the (potentially memory-mapped) region of the buffer. The encoded frame data is then
  * copied to this buffer, and finish_output_buffer() is called. This function can be used to inform the
- * caller tha the encoder is done with this buffer; it now contains encoded data, and will not be modified
+ * caller that the encoder is done with this buffer; it now contains encoded data, and will not be modified
  * further. encoded_frame is filled with information about the encoded frame data.
  * If acquiring the buffer fails, acquire_output_buffer() returns a NULL pointer.
  * NOTE: again, finish_output_buffer() is NOT a function to free the buffer; it just signals that the encoder
@@ -1535,6 +1642,12 @@ void imx_vpu_enc_configure_intra_qp(ImxVpuEncoder *encoder, int intra_qp);
  * even if an error occurs, unless the error occurred before the acquire_output_buffer() call, in which case
  * the encoded_frame's acquired_handle field will be set to NULL.
  *
+ * The aforementioned sequences involve a copy (encoded data is copied into the acquired buffer). As an
+ * alternative, a write-callback-style mode of operation can be used. This alternative mode is active if
+ * the write_output_data function pointer in encoding_params is not NULL. In this mode, neither
+ * acquire_output_buffer() nor finish_output_buffer() are called. Instead, whenever the encoder needs to
+ * write out data, it calls write_output_data().
+ *
  * The other fields in encoding_params specify additional encoding parameters, which can vary from frame to
  * frame.
  * output_code is a bit mask containing information about the encoding result. The value is a bitwise OR
